#!/usr/bin/perl -w
#
# Test for deleting documents in an index.
#

use strict;
require Cwd;
use File::Copy;
require 'pltests.pl';

my $cwd = Cwd::getcwd();
my $LOG = "$cwd/test-log";
my $TARGET = pltests::selectdata();
my $INDEX = "$cwd/idx4";
my $TMPDATA = "$cwd/tmp-data";
my $MKNMZ = pltests::binpath('mknmz');

my $cmd;

pltests::putline($LOG, "  *** starting $0");

if (-d "$INDEX") {
    unlink(<$INDEX/NMZ.*>);
} else {
    mkdir("$INDEX");
}

# Copy docments.
if (-d "$TMPDATA") {
    unlink(<$TMPDATA/*>);
} else {
    mkdir("$TMPDATA");
}
for my $file (<$TARGET/*>) {
    copy($file, $TMPDATA);
}

# Simple indexing.
$cmd = "$MKNMZ -O $INDEX $TMPDATA";
pltests::putline($LOG, "  $cmd");
system("$cmd >> $LOG");
if ($? != 0) {
    exit 1;
}

my $prevdocnum = 0;
my $doc = pltests::cat("$INDEX/NMZ.status");
unless (defined $doc) {
    exit 1;
}
if ($doc =~ /^files (\d+)/m) {
    $prevdocnum = $1;
} else {
    exit 1;
}

# Remove some documents.
unlink(<$TMPDATA/*.txt>);

# Update the index.
$cmd = "$MKNMZ --update=$INDEX";
pltests::putline($LOG, "  $cmd");
system("$cmd >> $LOG");
if ($? != 0) {
    exit 1;
}

my $postdocnum = 0;
$doc = pltests::cat("$INDEX/NMZ.status");
unless (defined $doc) {
    exit 1;
}
if ($doc =~ /^files (\d+)/m) {
    $postdocnum = $1;
} else {
    exit 1;
}

pltests::putline($LOG, "predocnum: $prevdocnum postdocnum: $postdocnum");

if ($prevdocnum <= $postdocnum) {
    exit 1;
}

exit 0;
